/**************************************************************************
 * C/C++ Headers, Using Declarations, Typedefs
 **************************************************************************/

#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstdio>
#include <cctype>
#include <string>
#include <vector>
#include <map>

#ifdef HEADLESS

# include <windows.h>
# pragma  comment(lib, "kernel32.lib")
# pragma  comment(lib, "user32.lib")

#endif

using std::string;
using std::vector;
using std::ostringstream;
using std::map;

using std::cout;
using std::endl;

typedef map<string,string> StringMap;

/**************************************************************************
 * Local Functions
 **************************************************************************/

static void doShow(string s)
{
#ifdef HEADLESS

   const char *c = s.c_str();

   MessageBox(
      NULL, (char *)c, "Message", MB_OK | MB_ICONINFORMATION
   );

#else

   printf("%s\n", s.c_str());

#endif
}

// ------------------------
// begin: HEADLESS {
// ------------------------

#ifdef HEADLESS

struct RunInfo
{
   HANDLE mProcess;
   HANDLE mThread;

   RunInfo() : mProcess(0), mThread(0) {}
};

static string getSysErrMsg()
{
   LPVOID lpMsgBuf;

   FormatMessage( 
      FORMAT_MESSAGE_ALLOCATE_BUFFER | 
      FORMAT_MESSAGE_FROM_SYSTEM     | 
      FORMAT_MESSAGE_IGNORE_INSERTS,
      NULL,
      GetLastError(),
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
      (LPTSTR)&lpMsgBuf,
      0,
      NULL 
   );

   char *charPtr = (char *)lpMsgBuf;
   string msg    = charPtr;

   LocalFree(lpMsgBuf);

   return(msg);
}

static int doWait(bool debug, RunInfo &info)
{
   int returnValue = 0;

   if(debug)
   {
      doShow("begin - doWait()");
   }

   WaitForSingleObject(info.mProcess, INFINITE);

   unsigned long value = 0L;

   if(GetExitCodeProcess(info.mProcess, &value))
      returnValue = (int)value;

   CloseHandle(info.mProcess);
   CloseHandle(info.mThread);

   if(debug)
   {
      ostringstream oss;

      oss << "end - doWait() - returnValue: " << returnValue;

      doShow(oss.str());
   }

   return(returnValue);
}

static bool doRun(const string &theCommand, RunInfo &info, string &errMsg)
{
   STARTUPINFO          si = { sizeof(si) };
   PROCESS_INFORMATION  pi;

   const char *c = theCommand.c_str();

   BOOL status = CreateProcess(
      0,
      (char *)c,
      0,
      0,
      FALSE,
      0,
      0,
      0,
      &si,
      &pi
   );

   if(!status)
   {
      ostringstream oss;

      oss << "Error executing: '" << theCommand << "'";

      string sysErrMsg = getSysErrMsg();

      if(sysErrMsg != "")
         oss << endl << sysErrMsg;

      errMsg = oss.str();

      return(false);
   }

   info.mProcess = pi.hProcess;
   info.mThread  = pi.hThread;

   return(true);
}

static vector<string> getArgs()
{
   typedef enum {
      BEGIN  = 0,
      WORD   = 1,
      QUOTE  = 2,
      ESCAPE = 3
   } State;

   State state = BEGIN;

   string cmdLine = GetCommandLine();
   string current = "";

   vector<string> v;

   for(int n = cmdLine.length(), i = 0; i < n; i ++)
   {
      char c = cmdLine[i];

      switch(state)
      {
         case BEGIN:
         {
            if(c == ' ' || c == '\t')
            {
               // Remain in state 'BEGIN'
            }
            else if(c == '"')
            {
               state = QUOTE;
            }
            else
            {
               state    = WORD;
               current += c;
            }
         }
         break;

         case WORD:
         {
            if(c == ' ' || c == '\t' || c == '"')
            {
               v.push_back(current);

               state   = (c == '"' ? QUOTE : BEGIN);
               current = "";
            }
            else
            {
               current += c;
            }
         }
         break;

         case QUOTE:
         {
            if(c == '"')
            {
               v.push_back(current);

               current = "";
               state   = BEGIN;
            }
            else if(c == '\\')
            {
               state = ESCAPE;
            }
            else
            {
               current += c;
            }
         }
         break;

         case ESCAPE:
         {
            if(c != '"')
               current += '\\';

            current += c;
            state    = QUOTE;
         }
         break;
      }
   }

   // We can end in state 'WORD', but ending in state 'QUOTE'
   // or 'ESCAPE' indicates user error.  However, in the interests of
   // simplicity we just pick up anything lingering in 'current', no
   // matter the final state

   if(current != "")
      v.push_back(current);

   return(v);
}

#endif

// ------------------------
// end: HEADLESS }
// ------------------------

static int execute(bool debug, string theCommand)
{
   int status = 0;

   if(debug)
   {
      doShow(theCommand);
   }

#ifdef HEADLESS

   RunInfo info;
   string  errMsg;

   if(!doRun(theCommand, info, errMsg))
   {
      doShow(errMsg);
   }
   else
   {
      status = doWait(debug, info);
   }
 
#else

   status = system(theCommand.c_str());

#endif

   return(status);
}

static bool getLines(
   const char *fileName, vector<string> &lines, string &errMsg
)
{
   FILE *f = NULL;

   if((f = fopen(fileName, "rb")) == NULL)
   {
      ostringstream oss;

      oss << "Could not open file: [" << fileName << "]";

      errMsg = oss.str();

      return(false);
   }

   string currentLine = "";
   int c              = 0;

   while((c = fgetc(f)) != EOF)
   {
      char ch = (char)c;
      currentLine += ch;

      if(ch == '\n')
      {
         lines.push_back(currentLine);
         currentLine = "";
      }
   }

   fclose(f);

   if(currentLine.length() > 0)
      lines.push_back(currentLine);

   return(true);
}

static string toLowerCase(string s)
{
   string lower = "";

   for(int length = s.length(), i = 0; i < length; i ++)
      lower += tolower(s[i]);

   return(lower);
}

static bool isValidKey(string s)
{
   int n = s.length();

   if(n == 0)
      return(false);

   for(int i = 0; i < n; i ++)
   {
      char c = s[i];

      if(!isalpha(c) && c != '_')
      {
         if(i == 0 || !isdigit(c))
            return(false);
      }
   }

   return(true);
}

static string rtrim(string s)
{
   int length = s.length();
   int index  = -1;

   for(int i = 0; i < length; i ++)
      if(!isspace(s[i]))
         index = i;

   if(index == -1)
      return("");

   return(s.substr(0, index + 1));
}

static string ltrim(string s)
{
   int length = s.length();
   int index  = -1;

   for(int i = 0; index == -1 && i < length; i ++)
      if(!isspace(s[i]))
         index = i;

   if(index == -1)
      return("");

   return(s.substr(index));
}

static string trim(string s)
{
   return(rtrim(ltrim(s)));
}

static bool getMap(
   vector<string> &lines, StringMap &theMap, string &errMsg
)
{
   for(int n = lines.size(), i = 0; i < n; i ++)
   {
      // If this line is blank or a comment (first non-whitespace character
      // is '#') then we ignore it

      string line = lines[i];
      int index   = -1;
      int nl      = line.length();
      int j       = 0;

      for(j = 0; index == -1 && j < nl; j ++)
         if(!isspace(line[j]))
            index = j;

      if(index != -1 && line[index] != '#')
      {
         int eqIndex = -1;

         for(j = index; eqIndex == -1 && j < nl; j ++)
            if(line[j] == '=')
               eqIndex = j;

         ostringstream oss;

         oss << "line[" << (i + 1) << "], ";

         if(eqIndex == -1)
         {
            oss << "missing '=' (" << rtrim(line) << ")";
            errMsg = oss.str();
            return(false);
         }

         string key   = toLowerCase(trim(line.substr(0, eqIndex)));
         string value = trim(line.substr(eqIndex + 1));

         if(!isValidKey(key))
         {
            oss << "invalid key '" << key << "' (" << rtrim(line) << ")";
            errMsg = oss.str();
            return(false);
         }

         if(value.length() == 0)
         {
            oss << "invalid value '" << value << "' (" << rtrim(line) << ")";
            errMsg = oss.str();
            return(false);
         }

         StringMap::iterator it = theMap.find(key);

         if(it != theMap.end())
         {
            oss << "duplicate key '" << key << "' (" << rtrim(line) << ")";
            errMsg = oss.str();
            return(false);
         }

         theMap[key] = value;
      }
   }

   return(true);
}

template <class T>
static string toString(vector<T> v, string separator)
{
   ostringstream oss;

   for(int n = v.size(), i = 0; i < n; i ++)
      oss << (i == 0 ? "" : separator) << v[i];

   return(oss.str());
}

static bool toInteger(string s, int &iValue)
{
   const char *ptr = s.c_str();
   char *pEnd      = NULL;
   long int value  = strtol(ptr, &pEnd, 10);
                                                                                
   if(*pEnd != '\0')
      return(false);

   iValue = (int)value;

   return(true);
}

static bool toPositiveInteger(string s, int &iValue)
{
   if(!toInteger(s, iValue))
      return(false);

   return(iValue > 0);
}

static bool resolve(
   int                depth,
   StringMap          &theMap,
   vector<string>     &history,
   string             key,
   string             &value,
   vector<string>     args,
   string             &errMsg
)
{
   StringMap::iterator it = theMap.find(key);

   if(it == theMap.end())
   {
      if(key == "*")
      {
         value = toString(args, " ");
         return(true);
      }

      ostringstream oss;

      int theValue = 0;

      if(toPositiveInteger(key, theValue))
      {
         if(theValue > (int)args.size())
         {
            oss << "index: [" <<  key << "] too large";
            errMsg = oss.str();
            return(false);
         }

         value = args[theValue - 1];
         return(true);
      }

      oss << "No match for key: [" << key << "]";

      errMsg = oss.str();

      return(false);
   }

   history.push_back(key);

   value = it->second;

   while(true)
   {
      int n     = value.length();
      int index = -1;
      int i     = 0;

      for(i = 0; index == -1 && i < n; i ++)
         if(value[i] == '%')
            index = i;

      if(index == -1)
         break;

      int left  = index;
      int right = -1;

      for(i = left + 1; right == -1 && i < n; i ++)
         if(value[i] == '%')
            right = i;

      if(right == -1)
      {
         ostringstream oss;

         oss << "Odd number of %'s in value: [" << value << "]";

         errMsg = oss.str();

         return(false);
      }

      string theKey = toLowerCase(
         value.substr(left + 1, (right - left) - 1)
      );

      // Watch for circular references ...

      bool found = false;

      for(int nh = history.size(), j = 0; !found && j < nh; j ++)
         if(theKey == history[j])
            found = true;

      if(found)
      {
         ostringstream oss;

         oss <<
            "circular reference: " <<
            "[" << toString(history, " -> ") << " -> " << theKey << "]";

         errMsg = oss.str();

         return(false);
      }

      string theValue;

      if(!resolve(depth + 1, theMap, history, theKey, theValue, args, errMsg))
         return(false);

      history.erase(
         history.end() - 1, history.end()
      );

      value = value.substr(0, left) + theValue + value.substr(right + 1);
   }

   // Store the resolved value back to 'theMap' in case
   // we reference it again

   theMap[key] = value;

   return(true);
}

static bool processMap(
   StringMap &theMap, vector<string> args, string &errMsg
)
{
   StringMap::iterator it = theMap.begin();

   while(it != theMap.end())
   {
      string key   = it->first;
      string value = "";

      vector<string> history;

      if(!resolve(0, theMap, history, key, value, args, errMsg))
         return(false);

      theMap[key] = value;

      it++;
   }

   return(true);
}

static bool mapFromFile(
   const char     *fileName,
   StringMap      &theMap,
   vector<string> args,
   string         &errMsg
)
{
   vector<string> lines;

   if(!getLines(fileName, lines, errMsg))
      return(false);

   if(!getMap(lines, theMap, errMsg))
      return(false);

   if(!processMap(theMap, args, errMsg))
      return(false);

   return(true);
}

static void dumpMap(StringMap &theMap)
{
   StringMap::iterator it = theMap.begin();
   int longest                     = 0;

   while(it != theMap.end())
   {
      int length = it->first.length();

      if(length > longest)
         longest = length;

      it++;
   }

   ostringstream oss;

   for(it = theMap.begin(); it != theMap.end(); it ++)
   {
      if(it != theMap.begin())
         oss << endl;

      oss << it->first;

      int n = longest - it->first.length();

      for(int i = 0; i < n; i ++)
         oss << " ";

      oss << " => " << it->second;
   }

   doShow(oss.str());
}

bool getDebugMode()
{
   const char *ptr  = getenv("RUNNER_DEBUG");
   bool returnValue = false;

   if(ptr)
   {
      if(*ptr == 'y' || *ptr == 'Y' ||
         *ptr == 't' || *ptr == 'T' || *ptr == '1')
      {
         returnValue = true;
      }
   }

   return(returnValue);
}

/**************************************************************************
 * runnerMain
 **************************************************************************/

static int runnerMain(int argc, char **argv)
{
   // ------------------------------------
   // Expected command-line arguments
   //
   //    argv[0] => this exe's name
   //    argv[1] => file containing key/value pairs
   //    argv[2] => key to execute
   //
   // Any arguments after argv[2] will be put into
   // the vector 'args' for reference via: %1%, %2%, %*%
   //
   //    argv[3] -> args[0] -> %1%
   //    argv[4] -> args[1] -> %2%
   //
   //    argv[3] .. argv[n] -> %*%
   // ------------------------------------

   ostringstream oss;

   if(argc < 3)
   {
      oss << "Usage: " << argv[0] << " fileName keyName { ... args ... }";
      doShow(oss.str());
      return(1);
   }

   vector<string> args;

   for(int i = 3; i < argc; i ++)
      args.push_back(argv[i]);

   bool debug = getDebugMode();

   if(debug)
   {
      oss << "args: [";

      for(int j = 0; j < argc; j ++)
         oss << (j == 0 ? "" : ", ") << argv[j];

      oss << "]";

      doShow(oss.str());

      oss.str("");
   }

   const char *fileName = argv[1];
   const char *keyName  = argv[2];
   string errMsg        = "";

   StringMap theMap;

   if(!mapFromFile(fileName, theMap, args, errMsg))
   {
      doShow(errMsg);
      return(1);
   }

   if(debug)
   {
      dumpMap(theMap);
   }

   StringMap::iterator it = theMap.find(keyName);

   if(it == theMap.end())
   {
      oss << "Key not found: " << keyName;
      doShow(oss.str());
      return(1);
   }

   return(execute(debug, it->second));
}

#ifdef HEADLESS

int PASCAL WinMain(
   HINSTANCE hInstance,
   HINSTANCE hPrevInstance,
   LPSTR     lpCmdLine,
   int       nCmdShow
)
{
   vector<string> args = getArgs();
   int size            = args.size();
   char **argv         = new char *[size];

   for(int i = 0; i < size; i ++)
      argv[i] = (char *)args[i].c_str();

   int status = runnerMain(size, argv);

   delete [] argv;

   return(status);
}

#else

int main(int argc, char **argv)
{
   return(runnerMain(argc, argv));
}

#endif
